{*******************************************
 *             FuzzyReport 2.0             *
 *                                         *
 *  Copyright (c) 2000 by Fabio Dell'Aria  *
 *                                         *
 *-----------------------------------------*
 * For to use this source code, you must   *
 * read and agree all license conditions.  *
 *******************************************}

unit FR_Prev;

interface

{$I FR_Vers.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, Buttons, FR_Const, FR_Class, FR_Find, Menus
  {$IFDEF Delphi4}, ImgList{$ENDIF};

const
  CM_SetZoom  = WM_USER + 1;
  crMyCursor = 22;

type

  TZoomType= (ztNormal,ztWidth,ztOnePage,ztTwoPages);

  TPaintPanel = class(TPanel)
  private
    { Private declarations }
  public
    { Public declarations }
    NoPaint, Drag, NoSetY: Boolean;
    DragX, DragY: Integer;
    procedure Paint; override;
    procedure MouseUp (Button: TMouseButton; Shift: TShiftState;
                       X, Y: Integer); override;
    procedure MouseDown (Button: TMouseButton; Shift: TShiftState;
                         X, Y: Integer); override;
    procedure MouseMove (Shift: TShiftState; X, Y: Integer); override;
  end;

  TStandardPreview = class(TfrPreview)
    // Base Controls...
    StatusBar: TPanel;
    VPanel: TPanel;
    HPanel: TPanel;
    Bevel1: TBevel;
    Bevel2: TBevel;
    CloseBtn: TSpeedButton;
    OpenBtn: TSpeedButton;
    SaveBtn: TSpeedButton;
    PrintBtn: TSpeedButton;
    FindBtn: TSpeedButton;
    Panel0: TPanel;
    Panel2: TPanel;
    RightPanel: TPanel;
    ButtonsPanel: TPanel;
    PrevBtn: TSpeedButton;
    NextBtn: TSpeedButton;
    ZoomBtn: TComboBox;
    OpenDialog: TOpenDialog;
    SaveDialog: TSaveDialog;
    PrintDialog: TPrintDialog;
    Timer: TTimer;
    PopupMenu: TPopupMenu;
    ImageList: TImageList;
    LeftPanel: TPanel;
    PagePanel: TPanel;
    PaintPage: TPaintBox;
    PaintPage2: TPaintBox;
    ScreenBtn: TSpeedButton;
    Panel1: TPanel;
    VI_Panel: TPanel;
    VScroll: TScrollBar;
    HI_Panel: TPanel;
    HScroll: TScrollBar;
    // Base procedures...
    procedure FormCreate(Sender: TObject);
    procedure VScrollChange(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure HScrollChange(Sender: TObject);
    procedure CloseBtnClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure PrevBtnClick(Sender: TObject);
    procedure NextBtnClick(Sender: TObject);
    procedure ZoomBtnClick(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure OpenBtnClick(Sender: TObject);
    procedure SaveBtnClick(Sender: TObject);
    procedure PrintBtnClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure ZoomBtnKeyPress(Sender: TObject; var Key: Char);
    procedure FindBtnClick(Sender: TObject);
    procedure TimerTimer(Sender: TObject);
    procedure PaintPagePaint(Sender: TObject);
    procedure ScreenBtnClick(Sender: TObject);
    procedure ZoomBtnKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure HI_PanelResize(Sender: TObject);
    procedure VI_PanelResize(Sender: TObject);
  private
    { Private declarations }
    Panel: TPaintPanel;
    Vposition,HPosition,FPageNo: Integer;
    FZoom: Double;
    FindText, TextFound, CaseSensitive, SubString,
    NextItem, PrevItem, CurrPage, ResizePaint: Boolean;
    OldWindowState: TWindowState;
    TextToFind: String;
    ExtRect: TRect;
    PageFound, FoundTimer, ObjectNo, LastObject, CurrObject,
    MaxWidth, MaxHeight, Start, Len: Integer;
    LastY, LastZoom: Double;
    FZoomType, LastZoomType: TZoomType;
    procedure SetPageNumber (Value:Integer);
    procedure SetZoom (Value:Double);
    procedure CheckZoom;
    procedure SetZoomFactor;
    function  FindXY (Pn: Integer): TPoint;
    procedure MenuClick (Sender: TObject);
    procedure CMSetZoom (Var Message:TMessage); Message CM_SetZoom;
    {$IFDEF Delphi4}
    procedure MouseWheelUp   (Sender: TObject; Shift: TShiftState;
                              MousePos: TPoint; var Handled: Boolean);
    procedure MouseWheelDown (Sender: TObject; Shift: TShiftState;
                              MousePos: TPoint; var Handled: Boolean);
    {$ENDIF}
  protected
    // Protected procedures & data ...
    FEnableStatusBar, FEnableBigScreen, FEnablePagePanel, FOldEnableStatusBar: Boolean;
    FSmallChange, FLargeChange: Integer;
    procedure DrawPage (Cnv:HDC;N:Integer;R:TRect); override;
    procedure SetStatusBar (Value:Boolean);
    procedure SetBigScreen (Value:Boolean);
    procedure SetPagePanel (Value:Boolean);
    procedure SetSmallChange (Value:Integer);
    procedure SetLargeChange (Value:Integer);
  public
    { Public declarations }
    procedure Open; virtual;
    procedure Save; virtual;
    procedure Print; virtual;
    procedure Find; virtual;
    procedure First; virtual;
    procedure Last; virtual;
    procedure Next; virtual;
    procedure Prior; virtual;
    procedure ZoomWidth; virtual;
    procedure ZoomOnePage; virtual;
    procedure ZoomTwoPages; virtual;
    property EnableBigScreen :Boolean read FEnableBigScreen write SetBigScreen default False;
    property EnableStatusBar :Boolean read FEnableStatusBar write SetStatusBar default True;
    property EnablePagePanel :Boolean read FEnablePagePanel write SetPagePanel default True;
    property ZoomType:TZoomType read FZoomType;
    property PageNo  :Integer read FPageNo write SetPageNumber;
    property Zoom    :Double read FZoom write SetZoom;
    property SmallChange:Integer read FSmallChange write SetSmallChange default 10;
    property LargeChange:Integer read FLargeChange write SetLargeChange default 200;
  end;

implementation

{$R *.DFM}
{$R Hand.Res}

var WinNt: Boolean;

//------------------------------------------------------------------------------
procedure TPaintPanel.Paint;
Var I, Y, X, MaxY, MaxX, BigY, NewPageNo: Integer;
    Dx, Dy: Double;
    R: TRect;
Begin
  Canvas.Brush.Color := clWhite;
  Canvas.Pen.Color   := clBlack;
  Y    := 10 - TStandardPreview(Parent).VScroll.Position;
  MaxY := 10;
  NewPageNo := 1;
  BigY := 0;
  For I := 0 To TStandardPreview(Parent).Report.EMFPages.Count-1 Do
    With TStandardPreview(Parent).Report.EMFPages.Pages[I].Pict Do
      Begin
        Dx := (TStandardPreview(Parent).Report.EMFPages.Pages[I].pgWidth * mmToPixel) /
              TStandardPreview(Parent).Report.EMFPages.Pages[I].Pict.Width *
              TStandardPreview(Parent).Zoom / 100;
        Dy := (TStandardPreview(Parent).Report.EMFPages.Pages[I].pgHeight * mmToPixel) /
              TStandardPreview(Parent).Report.EMFPages.Pages[I].Pict.Height *
              TStandardPreview(Parent).Zoom / 100;
        If Not NoPaint Then
          Begin
            If Y<=10 Then NewPageNo := I + 1;
            If Not((Y>Canvas.ClipRect.Bottom) Or
               (Round(Y+Height*Dy+1)<Canvas.ClipRect.Top)) Then
              Begin
                If Round(Width*Dx-Self.Width)>0 Then X := 10
                                                Else
                  If TStandardPreview(Parent).ZoomType<>ztTwoPages Then
                    X := Round ((Self.Width-(Width*Dx+20)-2)/2)+10
                                                      Else
                  If I Mod 2=0 Then
                    X := Round ((Self.Width/2-(Width*Dx)-2)/2)
                               Else
                    X := Round ((Self.Width/2-(Width*Dx)-2)/2 + Self.Width/2);
                Dec (X,TStandardPreview(Parent).HScroll.Position);
                If Y<=10 Then TStandardPreview(Parent).LastY := Abs(Y-10)/(Height*Dy);
                R := Rect (X,Y,Round(X+Width*Dx+1),Round(Height*Dy+Y+1));
                Canvas.Rectangle (R.Left,R.Top,R.Right,R.Bottom);
                Canvas.Rectangle (R.Right,R.Top+3,R.Right+1,R.Bottom);
                Canvas.Rectangle (R.Left+3,R.Bottom,R.Right+1,R.Bottom+1);
                R.Right := R.Right - 1;
                R.Bottom := R.Bottom - 1;
                TStandardPreview(Parent).DrawPage (Canvas.Handle,I,R);
              End;
          End;
        If Round(Height*Dy)>BigY Then BigY := Round(Height*Dy);
        If (TStandardPreview(Parent).ZoomType<>ztTwoPages) Or (I Mod 2=1) Then
          Begin
            Inc (Y,BigY+12);
            Inc (MaxY,BigY+12);
            BigY := 0;
          End;
      End;
  If Not NoPaint Then
    Begin
      If (TStandardPreview(Parent).ZoomType=ztTwoPages) And
         (NewPageNo>1) Then Dec (NewPageNo);
      If NewPageNo<>TStandardPreview(Parent).PageNo Then
        Begin
          TStandardPreview(Parent).FPageNo := NewPageNo;
          TStandardPreview(Parent).PaintPagePaint (Nil);
        End;
    End;
  Dec (MaxY,Height);
  If MaxY>0 Then Begin
                   TStandardPreview(Parent).VScroll.Max := MaxY;
                   TStandardPreview(Parent).VScroll.Enabled := True;
                 End
            Else Begin
                   TStandardPreview(Parent).VScroll.Max := 0;
                   TStandardPreview(Parent).VScroll.Enabled := False;
                 End;
  MaxX := Round (TStandardPreview(Parent).MaxWidth * TStandardPreview(Parent).Zoom / 100);
  If MaxX - Width >0 Then Begin
                            TStandardPreview(Parent).HScroll.Max := MaxX - Width + 20;
                            TStandardPreview(Parent).HScroll.Enabled := True;
                          End
                     Else Begin
                            TStandardPreview(Parent).HScroll.Max := 0;
                            TStandardPreview(Parent).HScroll.Enabled := False;
                          End;
End;

procedure TPaintPanel.MouseUp (Button: TMouseButton; Shift: TShiftState;
                               X, Y: Integer);
Var P: TPoint;
    I: Integer;
begin
  If Drag Then Begin
                 Drag  := False;
               End;
  With TStandardPreview(Parent) Do
    Begin
      ZoomBtnClick (Nil);
      If Button=mbRight Then
        Begin
          P := Panel.ClientToScreen (Point(X,Y));
          For I := 0 To PopupMenu.Items.Count-1 Do
            If ((ZoomType=ztNormal) And
                (PopupMenu.Items[I].Caption=ZoomBtn.Text)) Or
               ((ZoomType=ztWidth) And
                (PopupMenu.Items[I].Caption=FRConst_ZoomWidth)) Or
               ((ZoomType=ztOnePage) And
                (PopupMenu.Items[I].Caption=FRConst_ZoomOnePage)) Or
               ((ZoomType=ztTwoPages) And
                (PopupMenu.Items[I].Caption=FRConst_ZoomTwoPages)) Then
              PopupMenu.Items[I].Checked := True
                                                        Else
              PopupMenu.Items[I].Checked := False;
          PopupMenu.Popup (P.X,P.Y);
        End;
    End;
end;

procedure TPaintPanel.MouseDown (Button: TMouseButton; Shift: TShiftState;
                                 X, Y: Integer);
begin
  If Button=mbLeft Then Begin
                          Drag   := True;
                          DragX  := X;
                          DragY  := Y;
                        End;
end;

procedure TPaintPanel.MouseMove (Shift: TShiftState; X, Y: Integer);
begin
  If Drag Then
    With TStandardPreview(Parent) Do
    Begin
      VScroll.Position := Vscroll.Position - (Y - DragY);
      HScroll.Position := HScroll.Position - (X - DragX);
      DragX := X;
      DragY := Y;
    End;
end;
//------------------------------------------------------------------------------
procedure TStandardPreview.SetPageNumber (Value:Integer);
Var Y, I, BigY: Integer;
    D: Double;
begin
  BigY := 0;
  If (Value>=1) And (Value<=Report.EMFPages.Count) Then
    Begin
      Y := 0;
      For I := 0 To Value-2 Do
        Begin
          D := (Report.EMFPages.Pages[I].pgHeight * mmToPixel) /
               Report.EMFPages.Pages[I].Pict.Height * Zoom / 100;
          If Round(Report.EMFPages.Pages[I].Pict.Height*D)>BigY Then
            BigY := Round(Report.EMFPages.Pages[I].Pict.Height*D);
          If (ZoomType<>ztTwoPages) Or (I Mod 2=1) Then
            Begin
              Inc (Y,BigY+12);
              BigY := 0;
            End;
        End;
      If VScroll.Position<>Y Then VScroll.Position := Y;
      FPageNo := Value;
      PaintPagePaint (Nil);
    End;
end;

procedure TStandardPreview.SetZoom (Value:Double);
Var // Dy: Double;
    R: TRect;
    S: TNotifyEvent;
begin
  If (Value<>FZoom) And (TextFound) Then
    Begin
      ExtRect.Left   := Round(ExtRect.Left * (Value/FZoom));
      ExtRect.Top    := Round(ExtRect.Top * (Value/FZoom));
      ExtRect.Right  := Round(ExtRect.Right * (Value/FZoom));
      ExtRect.Bottom := Round(ExtRect.Bottom * (Value/FZoom));
    End;
  Panel.NoPaint := True;
  FZoom := Value;
  Panel.Paint;
  If (Not Panel.NoSetY) And (IsWindowVisible (Handle)) Then
    Begin
      SetPageNumber (PageNo);
      S := VScroll.OnChange;
      VScroll.OnChange := Nil;
(*      Dy := (Report.EMFPages.Pages[PageNo-1].pgHeight * mmToPixel) /
            Report.EMFPages.Pages[PageNo-1].Pict.Height * Zoom / 100;
      VScroll.Position := VScroll.Position +
                          Round(LastY * Report.EMFPages[PageNo-1].Pict.Height * Dy);*)
      VScroll.OnChange := S;
      VPosition := VScroll.Position;
    End;
  PostMessage (Handle,CM_SetZoom,0,0);
  If HScroll.Max<=70 Then HScroll.Position := HScroll.Max Div 2
                     Else HScroll.Position := 0;
  Panel.NoPaint := False;
  If (LastZoom<>Zoom) Or (LastZoomType<>ZoomType) Or (ResizePaint) Then
    Begin
      R := Rect (0,0,Panel.Width,Panel.Height);
      InvalidateRect (Panel.Canvas.Handle,@R,False);
      Panel.Canvas.Brush.Color := Panel.Color;
      Panel.Canvas.FillRect (Rect(0,0,Panel.Width,Panel.Height));
      Panel.Paint;
      LastZoom     := Zoom;
      LastZoomType := ZoomType;
      ResizePaint := False;
    End;
end;

procedure TStandardPreview.CheckZoom;
Var S: String;
    I: Double;
    W: Integer;
begin
  If Screen.ActiveControl=ZoomBtn Then
    Begin
      ZoomBtn.Enabled := False;
      S := ZoomBtn.Text;
      If S=FRConst_ZoomWidth Then Begin
                                    FZoomType := ztWidth;
                                    SetZoomFactor;
                                  End
                             Else
      If S=FRConst_ZoomOnePage Then Begin
                                      FZoomType := ztOnePage;
                                      SetZoomFactor;
                                    End
                               Else
      If S=FRConst_ZoomTwoPages Then Begin
                                       FZoomType := ztTwoPages;
                                       SetZoomFactor;
                                     End
                                Else
        Begin
          If Pos('%',S)=Length(S) Then Delete (S,Pos('%',S),1);
          If Pos(DecimalSeparator,S)>0 Then S[Pos(DecimalSeparator,S)]:='.';
          Val (S,I,W);
          If (W<>0) Or (I=0) Then Begin
                                    Beep;
                                    PostMessage (Handle,CM_SetZoom,0,0);
                                  End
                             Else If (ZoomType<>ztNormal) Or (Zoom<>I) Then
                                    Begin
                                      FZoomType := ztNormal;
                                      Zoom := I;
                                    End
                                                                       Else
                                    PostMessage (Handle,CM_SetZoom,0,0);
        End;
      ZoomBtn.Enabled := True;
    End;
end;

procedure TStandardPreview.SetZoomFactor;
Var Z: Double;
begin
  Case ZoomType of
    ztWidth   : Zoom := (Panel.Width - 20) / MaxWidth * 100;
    ztOnePage : Begin
                  Z := (Panel.Width - 20) / MaxWidth;
                  If MaxHeight*Z>Panel.Height-20 Then
                    Z := Z * ((Panel.Height - 20) / (MaxHeight*Z));
                  Zoom := Z * 100;
                End;
    ztTwoPages: Begin
                  Z := (Panel.Width - 20) / (MaxWidth * 2);
                  If MaxHeight*Z>Panel.Height-20 Then
                    Z := Z * ((Panel.Height - 20) / (MaxHeight*Z));
                  Zoom := Z * 100;
                End;
    ztNormal  : Zoom := Zoom;
  End;
end;

function TStandardPreview.FindXY (Pn: Integer): TPoint;
Var I, Y, X, BigY: Integer;
    Dx, Dy: Double;
Begin
  Y := 10 - VScroll.Position;
  I := 0;
  BigY := 0;
  While I<=Pn Do
    With Report.EMFPages.Pages[I].Pict Do
      Begin
        If I=Pn Then
          Begin
            Dx := (Report.EMFPages.Pages[I].pgWidth * mmToPixel) /
                  Report.EMFPages.Pages[I].Pict.Width * Zoom / 100;
            If Round(Width*Dx-Panel.Width)>0 Then X := 10
                                             Else
            If ZoomType<>ztTwoPages Then
              X := Round ((Panel.Width-(Width*Dx+20)-2)/2)+10
                                    Else
            If I Mod 2=0 Then
              X := Round ((Panel.Width/2-(Width*Dx)-2)/2)
                         Else
              X := Round ((Panel.Width/2-(Width*Dx)-2)/2 + Panel.Width/2);
            Dec (X,HScroll.Position);
            Result := Point (X,Y);
          End;
        If I<Pn Then
          Begin
            Dy := (Report.EMFPages.Pages[I].pgHeight * mmToPixel) /
                  Report.EMFPages.Pages[I].Pict.Height * Zoom / 100;
            If Round(Height*Dy)>BigY Then BigY := Round(Height*Dy);
            If (ZoomType<>ztTwoPages) Or (I Mod 2=1) Then
              Begin
                Inc (Y,BigY+12);
                BigY := 0;
              End;
          End;
        Inc (I);
      End;
end;


procedure TStandardPreview.DrawPage (Cnv:HDC;N:Integer; R:TRect);
const EMFDraw:Set Of Byte =[2,3,4,5,6,7,8,15,41,42,43,44,45,46,47,53,
                            55,56,71,72,73,74,76,77,78,79,81,83,84,85,
                            86,87,88,89,90,91,92,96,97];
      Chars  : Set Of Char=[' ','.',';',':','?','!','''','(',')','[',']','"'];
Type Params = Record
                Pn, ObjNo: Integer;
                Rct: TRect;
                Dx, Dy, LastAngle: Double;
                NewDraw: Boolean;
                Angs: TStringList;
                StdPreview: TStandardPreview;
                WDx,WDy,Wx,Wy,VDx,VDy,Vx,Vy: Integer;
                DrawAll: Boolean;
              End;
     PParams = ^ Params;
     Arr=Array [0..999] Of Integer;
     PArr=^ Arr;

     SaveFont = Record
                  Angle: Double;
                End;
     PSaveFont = ^SaveFont;

Var Param: PParams;

  function CustomEnhMetaFileProc (DC:HDC; HandleTable:PHandleTable;
                                  MyEnhMetaRecord:PEnhMetaRecord;
                                  nObj:Integer; LParam:Pointer):Bool; stdcall;
  Var E: TEMRExtTextOut;
      Q: PEMRExtCreateFontIndirect;
      Image: Pointer;
      Info: ^ TBitmapInfo;
      B: PEMRStretchBlt;
      X: PEMRBitBlt;
      Y: PEMRSetViewPortExtEx;
      Z: PEMRSetViewPortOrgEx;
      D: PEMRDeleteObject;
      P: TLogPen;
      S, S0, Text: String;
      I, OldBKMode, OldMapMode: Integer;
      Wnd, Vprt: TSize;
      Wp, Vp: TPoint;
      Pnt: TPoint;
      Metric: TTextMetric;
      OldFont: HFont;
      OldTxtColor, OldBKColor: TColor;
      DrawOK: Boolean;
      TheParam: PParams;
      SF: PSaveFont;

  function RectOK (R1, R2: TRect): Boolean;
  Var P: TLogPen;
  begin
    InflateRect (R2,1,1);
    GetObject (GetCurrentObject(DC,OBJ_PEN),SizeOf(P),@P);
    InflateRect (R1,4+P.lopnWidth.x Div 2,4+P.lopnWidth.x Div 2);
    R1.Left   := Round(R1.Left * TheParam^.Dx);
    R1.Right  := Round(R1.Right * TheParam^.Dx);
    R1.Top    := Round(R1.Top * TheParam^.Dy);
    R1.Bottom := Round(R1.Bottom * TheParam^.Dy);
    OffsetRect (R1,TheParam^.Rct.Left,TheParam^.Rct.Top);
    LpToDP (DC,R2,2);
    Result := Not ((R1.Right<R2.Left) Or (R1.Left>R2.Right) Or
                   (R1.Top>R2.Bottom) Or (R1.Bottom<R2.Top));
  end;

  procedure NewTextOut;
  Var Sw: PWideChar;
      I, D, X, N: Integer;
      C, V: TSize;
      A: PArr;
      Sx, Sy, Sy0: Double;
      W: Arr;
  begin
    If (TheParam^.NewDraw) And
       (TheParam^.WDx=1) And (TheParam^.WDy=1) And
       (TheParam^.Wx=0) And (TheParam^.Wy=0) And
       (TheParam^.VDx=1) And (TheParam^.VDy=1) And
       (TheParam^.Vx=0) And (TheParam^.Vy=0) Then
      Begin
        OldMapMode := SetMapMode (DC,MM_ANISOTROPIC);
        If Not WinNT Then
          Begin
            SetWindowExtEx (DC,1000,1000,@Wnd);
            SetWindowOrgEx (DC,0,0,@Wp);
          End
                     Else
          Begin
            SetWindowExtEx (DC,Round(TheParam^.Dx*1000),Round(TheParam^.Dy*1000),@Wnd);
            SetWindowOrgEx (DC,TheParam^.Rct.Left,TheParam^.Rct.Top,@Wp);
          End;
        SetViewPortExtEx (DC,1000,1000,@Vprt);
        SetViewPortOrgEx (DC,TheParam^.Rct.Left,TheParam^.Rct.Top,@Vp);
        E.emrtext.rcl.Left   := Round(E.emrtext.rcl.Left*TheParam^.Dx);
        E.emrtext.rcl.Top    := Round(E.emrtext.rcl.Top*TheParam^.Dy);
        E.emrtext.rcl.Right  := Round(E.emrtext.rcl.Right*TheParam^.Dx);
        E.emrtext.rcl.Bottom := Round(E.emrtext.rcl.Bottom*TheParam^.Dy);
        A := PArr(PChar(MyEnhMetaRecord)+E.EMRText.offDx);
        X := 0;
        N := 0;
        D := E.rclBounds.Right-E.rclBounds.Left+1;
        If TheParam^.LastAngle<>0 Then
          D := Round (Cos(TheParam^.LastAngle)*D+
               Sin(TheParam^.LastAngle)*(E.rclBounds.Bottom-E.rclBounds.Top+1));
        Repeat
          Inc (X,A^[N]);
          Inc (N);
        Until (X>=D) Or (N>=Round(E.emrtext.nChars));
        If N>=Round(E.emrtext.nChars) Then N:=Round(E.emrtext.nChars);
        Sw := PWideChar(PChar(MyEnhMetaRecord)+E.EMRText.offString);
        GetTextExtentPoint32W (DC,PWideChar(PChar(MyEnhMetaRecord)+
                                  E.EMRText.offString+(N-1)*2),1,V);
        GetTextExtentPoint32W (DC,Sw,N-1,C);
        If N>1 Then
          Begin
            If C.cx=0 Then C.Cx := 1;
            Sx := (X*TheParam^.Dx-V.cx)/C.cx
          End
               Else Sx := 1;
        Sy  := 0;
        For I := 0 To N-1 Do
          Begin
            GetTextExtentPoint32W (DC,PWideChar(PChar(MyEnhMetaRecord)+
                                   E.EMRText.offString+I*2),1,C);
            Sy0 := Sy;
            Sy  := Sy + C.cx;
            W[I] := Round(Sy*Sx) - Round(Sy0*Sx);
          End;
        ExtTextOutW (DC,Round(E.emrtext.ptlReference.x*TheParam^.Dx),
                        Round(E.emrtext.ptlReference.y*TheParam^.Dy),
                        E.emrtext.fOptions,@E.emrtext.rcl,
                        Sw,N,@W);
        SetMapMode (DC,OldMapMode);
        SetWindowExtEx (DC,Wnd.cx,Wnd.cy,Nil);
        SetWindowOrgEx (DC,Wp.x,Wp.Y,Nil);
        SetViewPortExtEx (DC,Vprt.cx,Vprt.cy,Nil);
        SetViewPortOrgEx (DC,Vp.x,Vp.y,Nil);
        DrawOK := False;
      End
                                             Else
    PlayEnhMetafileRecord (DC,HandleTable^,MyEnhMetaRecord^,nObj);
  end;

  begin
    Result := True;
    TheParam := PParams(LParam);
    If Not TheParam^.StdPreview.FindText Then
      Begin
        If MyEnhMetaRecord^.iType=1 Then TheParam^.ObjNo := 10000 * TheParam^.Pn
                                    Else Inc (TheParam^.ObjNo);
        If TheParam^.NewDraw Then
          Begin
            If (MyEnhMetaRecord^.iType=37) And (TheParam^.Angs.Count>0) Then
              Begin
                D := PEMRDeleteObject(MyEnhMetaRecord);
                If TheParam^.Angs.Find(IntToStr(D^.ihObject),I) Then
                  TheParam^.LastAngle := PSaveFont(TheParam^.Angs.Objects[I])^.Angle;
              End;
            If (MyEnhMetaRecord^.iType=40) And (TheParam^.Angs.Count>0) Then
              Begin
                D := PEMRDeleteObject(MyEnhMetaRecord);
                If TheParam^.Angs.Find(IntToStr(D^.ihObject),I) Then
                  Begin
                    Dispose (PSaveFont(TheParam^.Angs.Objects[I]));
                    TheParam^.Angs.Delete (I);
                    If TheParam^.Angs.Count=0 Then TheParam^.LastAngle := 0;
                  End;
              End;
            If MyEnhMetaRecord^.iType in [9,11] Then
              Begin
                Y := PEMRSetViewPortExtEx(MyEnhMetaRecord);
                If MyEnhMetaRecord^.iType=9 Then Begin
                                                   TheParam^.WDx := Y^.szlExtent.cx;
                                                   TheParam^.WDy := Y^.szlExtent.cy;
                                                 End
                                            Else Begin
                                                   TheParam^.VDx := Y^.szlExtent.cx;
                                                   TheParam^.VDy := Y^.szlExtent.cy;
                                                 End;
              End;
            If MyEnhMetaRecord^.iType in [10,12] Then
              Begin
                Z := PEMRSetViewPortOrgEx(MyEnhMetaRecord);
                If MyEnhMetaRecord^.iType=10 Then Begin
                                                    TheParam^.WDx := Z^.ptlOrigin.x;
                                                    TheParam^.WDy := Z^.ptlOrigin.y;
                                                  End
                                             Else Begin
                                                    TheParam^.VDx := Z^.ptlOrigin.x;
                                                    TheParam^.VDy := Z^.ptlOrigin.y;
                                                  End;
              End;
          End;
        If MyEnhMetaRecord^.iType=76 Then
          Begin
            X := PEMRBitBlt (MyEnhMetaRecord);
            If X^.cbBitsSrc>0 Then
              Begin
                Image := Pointer(PChar(MyEnhMetaRecord)+X^.offBitsSrc);
                Info  := Pointer(PChar(MyEnhMetaRecord)+X^.offBmiSrc);
                StretchDIBits (DC,X^.xDest,X^.yDest,X^.cxDest,X^.cyDest,
                                  X^.Xsrc,X^.ySrc,X^.cxDest,X^.cyDest,Image,Info^,
                                  X^.iUsageSrc,X^.dwRop);
                DrawOK := False;
              End;
          End
                                     Else
        If MyEnhMetaRecord^.iType=77 Then
          Begin
            B := PEMRStretchBlt (MyEnhMetaRecord);
            If B^.cbBitsSrc>0 Then
              Begin
                Image := Pointer(PChar(MyEnhMetaRecord)+B^.offBitsSrc);
                Info  := Pointer(PChar(MyEnhMetaRecord)+B^.offBmiSrc);
                StretchDIBits (DC,B^.xDest,B^.yDest,B^.cxDest,B^.cyDest,
                                  B^.Xsrc,B^.ySrc,B^.cxSrc,B^.cySrc,Image,Info^,
                                  B^.iUsageSrc,B^.dwRop);
                DrawOK := False;
              End;
          End
                                     Else
        If MyEnhMetaRecord^.iType=82 Then
          If (TheParam^.NewDraw) And
             (TheParam^.WDx=1) And (TheParam^.WDy=1) And
             (TheParam^.Wx=0) And (TheParam^.Wy=0) And
             (TheParam^.VDx=1) And (TheParam^.VDy=1) And
             (TheParam^.Vx=0) And (TheParam^.Vy=0) Then
            Begin
              Q := PEMRExtCreateFontIndirect(MyEnhMetaRecord);
              Pnt := Point (Q^.elfw.elfLogFont.lfWidth,Q^.elfw.elfLogFont.lfHeight);
              If (Q^.elfw.elfLogFont.lfEscapement<>0) And
                 (Q^.elfw.elfLogFont.lfOrientation<>0) Then
                Begin
                  New (SF);
                  SF^.Angle := (Q^.elfw.elfLogFont.lfOrientation / 10) / (2 * Pi);
                  TheParam^.Angs.AddObject (IntToStr(Q^.ihFont),TObject(SF));
                End;
              If Round(TheParam^.Dx*100)<>Round(TheParam^.Dy*100) Then
                Begin
                  Q^.elfw.elfLogFont.lfHeight :=
                    Round(Q^.elfw.elfLogFont.lfHeight * TheParam^.Dy * 10);
                  PlayEnhMetafileRecord (DC,HandleTable^,MyEnhMetaRecord^,nObj);
                  OldFont := SelectObject (DC,HandleTable^.objectHandle[Q^.ihFont]);
                  GetTextMetrics (DC,Metric);
                  Q^.elfw.elfLogFont.lfHeight := Round(Pnt.y * TheParam^.Dy);
                  Q^.elfw.elfLogFont.lfWidth  := Round(Metric.tmAveCharWidth / 10 *
                                                       (TheParam^.Dx/TheParam^.Dy));
                  SelectObject (DC,OldFont);
                  DeleteObject (HandleTable^.objectHandle[Q^.ihFont]);
                End
                                                                  Else
                Q^.elfw.elfLogFont.lfHeight :=
                  Round(Q^.elfw.elfLogFont.lfHeight * TheParam^.Dy);
              PlayEnhMetafileRecord (DC,HandleTable^,MyEnhMetaRecord^,nObj);
              Q^.elfw.elfLogFont.lfWidth  := Pnt.x;
              Q^.elfw.elfLogFont.lfHeight := Pnt.y;
              DrawOK := False;
            End
                               Else DrawOK := True
                                     Else
        If MyEnhMetaRecord^.iType in EMFDraw Then
          Begin
            E := PEMRExtTextOut(MyEnhMetaRecord)^;
            If Not TheParam^.DrawAll Then
              Begin
                DrawOK := RectOK (E.rclBounds,TheParam^.StdPreview.Panel.Canvas.ClipRect);
                If DrawOK Then
                  Begin
                    If (TheParam^.StdPreview.TextFound) And
                       (TheParam^.Pn=TheParam^.StdPreview.PageFound) And
                       (TheParam^.ObjNo=TheParam^.StdPreview.ObjectNo) Then
                      Begin
                        If TheParam^.StdPreview.FoundTimer Mod 2=1 Then
                          With TheParam^.StdPreview.Panel.Canvas Do
                            Begin
                              OldTxtColor := GetTextColor (DC);
                              OldBKColor  := GetBKColor(DC);
                              OldBKMode   := GetBKMode(DC);
                              If OldTxtColor=clWhite Then
                                SetTextColor (DC,clYellow Xor Round(OldTxtColor))
                                                     Else
                                SetTextColor (DC,clWhite);
                              SetBKColor (DC,clBlack);
                              SetBKMode (DC,OPAQUE);
                              NewTextOut;
                              SetBKColor (DC,OldBKColor);
                              SetBKMode (DC,OldBKMode);
                              SetTextColor (DC,OldTxtColor);
                              Rectangle (-10000,-10000,-10001,-10001);
                              DrawOK := False;
                            End;
                      End;
                  End;
              End
                                     Else DrawOK := True;
          End
                                                                      Else
          DrawOK := True;
        If DrawOK Then
          If MyEnhMetaRecord^.iType in [83,84] Then NewTextOut
                                               Else
            PlayEnhMetafileRecord (DC,HandleTable^,MyEnhMetaRecord^,nObj);
      End
                                Else
    If Not TheParam^.DrawAll Then
      If Not TheParam^.StdPreview.TextFound Then
        Begin
          If MyEnhMetaRecord^.iType=1 Then TheParam^.StdPreview.ObjectNo := 10000 * TheParam^.Pn
                                      Else
            TheParam^.StdPreview.ObjectNo := TheParam^.StdPreview.ObjectNo + 1;
          If MyEnhMetaRecord^.iType in [83,84] Then
            Begin
              Text := TheParam^.StdPreview.TextToFind;
              E    := PEMRExtTextOut(MyEnhMetaRecord)^;
              S    := WideCharToString(PWideChar(PChar(MyEnhMetaRecord) +
                                       E.EMRText.offString));
              S0   := S;
              I := 1;
              While I<=Length(S) Do
                If Copy(S,I,1)<' ' Then Delete (S,I,1)
                                   Else Inc (I);
              If Not TheParam^.StdPreview.CaseSensitive Then Begin
                                                               S    := UpperCase (S);
                                                               Text := UpperCase (Text);
                                                             End;
              TheParam^.StdPreview.TextFound := Pos(Text,S)>0;
              If (TheParam^.StdPreview.TextFound) And (Not TheParam^.StdPreview.SubString) Then
                TheParam^.StdPreview.TextFound := (Text=S) Or (((Pos(Text,S)=1) Or
                  (S[Pos(Text,S)-1] In Chars)) And
                  ((Pos(Text,S)+Length(Text)-1=Length(S)) Or
                  (S[Pos(Text,S)+Length(Text)] In Chars)));
              If TheParam^.StdPreview.TextFound Then
                TheParam^.StdPreview.CurrObject := TheParam^.StdPreview.CurrObject + 1;
              If (TheParam^.StdPreview.TextFound) And
                 (((Not TheParam^.StdPreview.NextItem) And (Not TheParam^.StdPreview.PrevItem) And
                 ((Not TheParam^.StdPreview.CurrPage) Or (TheParam^.Pn>=TheParam^.StdPreview.PageNo-1))) Or
                 ((TheParam^.StdPreview.NextItem) And
                 (TheParam^.StdPreview.CurrObject=TheParam^.StdPreview.LastObject+1)) Or
                 ((TheParam^.StdPreview.PrevItem) And
                 (TheParam^.StdPreview.CurrObject=TheParam^.StdPreview.LastObject-1))) Then
                Begin
                  TheParam^.StdPreview.LastObject := TheParam^.StdPreview.CurrObject;
                  TheParam^.StdPreview.ExtRect := E.rclBounds;
                  GetObject (GetCurrentObject(DC,OBJ_PEN),SizeOf(P),@P);
                  InflateRect (TheParam^.StdPreview.ExtRect,4+P.lopnWidth.x Div 2,
                               4+P.lopnWidth.x Div 2);
                  TheParam^.StdPreview.ExtRect.Left   := Round(TheParam^.StdPreview.ExtRect.Left * TheParam^.Dx);
                  TheParam^.StdPreview.ExtRect.Right  := Round(TheParam^.StdPreview.ExtRect.Right * TheParam^.Dx);
                  TheParam^.StdPreview.ExtRect.Top    := Round(TheParam^.StdPreview.ExtRect.Top * TheParam^.Dy);
                  TheParam^.StdPreview.ExtRect.Bottom := Round(TheParam^.StdPreview.ExtRect.Bottom * TheParam^.Dy);
                  TheParam^.StdPreview.PageFound      := TheParam^.Pn;
                  TheParam^.StdPreview.FoundTimer    := 0;
                  TheParam^.StdPreview.Timer.Enabled := True;
                  TheParam^.StdPreview.Start := Pos(Text,S);
                  TheParam^.StdPreview.Len   := Length(Text);
                  Result := False;
                End
                                                                Else
                TheParam^.StdPreview.TextFound := False;
            End;
        End;
  end;

begin
  New (Param);
  Param^.Pn := N;
  Param^.Rct := R;
  Param^.Dx := (Report.EMFPages.Pages[N].pgWidth * mmToPixel) /
                Report.EMFPages.Pages[N].Pict.Width * Zoom / 100;
  Param^.Dy := (Report.EMFPages.Pages[N].pgHeight * mmToPixel) /
                Report.EMFPages.Pages[N].Pict.Height * Zoom / 100;
  Param^.NewDraw := False;
  (*Not (((Round(Param^.Dx*100)=100) And (Round(Param^.Dy*100)=100)) Or
                    ((Round(Param^.Dx*100)<=75) And (Round(Param^.Dy*100)<=75)));*)
  Param^.Angs := TStringList.Create;
  Param^.Angs.Sorted := True;
  Param^.LastAngle := 0;
  Param^.StdPreview := Self;
  Param^.WDx := 1;
  Param^.WDy := 1;
  Param^.Wx  := 0;
  Param^.Wy  := 0;
  Param^.VDx := 1;
  Param^.VDy := 1;
  Param^.Vx  := 0;
  Param^.Vy  := 0;
  Param^.DrawAll := Cnv<>Panel.Canvas.Handle;
  EnumEnhMetafile (Cnv,Report.EMFPages.Pages[N].Pict.Handle,
                   @CustomEnhMetaFileProc,Param,R);
  Param^.Angs.Free;
  Dispose (Param);
end;

procedure TStandardPreview.MenuClick (Sender: TObject);
begin
  ZoomBtn.Text := TMenuItem(Sender).Caption;
  ZoomBtn.SetFocus;
  CheckZoom;
end;

procedure TStandardPreview.CMSetZoom (Var Message:TMessage);
Var S: String;
begin
  S := FloatToStr(Int(Zoom) + Round(Frac(Zoom)*10) / 10);
  ZoomBtn.Text := S + '%';
end;

{$IFDEF Delphi4}
procedure TStandardPreview.MouseWheelUp (Sender: TObject; Shift: TShiftState;
                                         MousePos: TPoint; var Handled: Boolean);
begin
  VScroll.Position := VScroll.Position - VScroll.SmallChange;
end;

procedure TStandardPreview.MouseWheelDown (Sender: TObject; Shift: TShiftState;
                                           MousePos: TPoint; var Handled: Boolean);
begin
  VScroll.Position := VScroll.Position + VScroll.SmallChange;
end;
{$ENDIF}

procedure TStandardPreview.FormCreate(Sender: TObject);

  procedure AddMenu(T: TMenuItem; B: TSpeedButton);
  Var M: TMenuItem;
  begin
    M := TMenuItem.Create (Nil);
    {$IFDEF Delphi4}
      M.ImageIndex := ImageList.AddMasked (B.Glyph,B.Glyph.Canvas.Pixels[0,15]);
    {$ENDIF}
    M.Caption := B.Caption;
    M.OnClick := B.OnClick;
    T.Add (M);
  end;

Var P: TMenuItem;
    I: Integer;
begin
  FEnablePagePanel  := True;
  FEnableStatusBar  := True;
  FEnableStatusBar  := True;
  FSmallChange      := 10;
  FLargeChange      := 200;
  Caption           := FRConst_PreviewCaption;
  OpenBtn.Caption   := FRConst_OpenBtnCaption;
  OpenBtn.Hint      := FRConst_OpenBtnHint;
  SaveBtn.Caption   := FRConst_SaveBtnCaption;
  SaveBtn.Hint      := FRConst_SaveBtnHint;
  PrintBtn.Caption  := FRConst_PrintBtnCaption;
  PrintBtn.Hint     := FRConst_PrintBtnHint;
  FindBtn.Caption   := FRConst_FindBtnCaption;
  FindBtn.Hint      := FRConst_FindBtnHint;
  ScreenBtn.Caption := FRConst_ScreenBtnCaption;
  ScreenBtn.Hint    := FRConst_ScreenBtnHint;
  CloseBtn.Caption  := FRConst_CloseBtnCaption;
  CloseBtn.Hint     := FRConst_CloseBtnHint;
  PrevBtn.Hint      := FRConst_PrevBtnHint;
  NextBtn.Hint      := FRConst_NextBtnHint;
  ZoomBtn.Hint      := FRConst_ZoomBtnHint;
  ZoomBtn.Items.Add (FRConst_ZoomWidth);
  ZoomBtn.Items.Add (FRConst_ZoomOnePage);
  ZoomBtn.Items.Add (FRConst_ZoomTwoPages);
  OpenDialog.Title  := FRConst_OpenDiatogTitle;
  OpenDialog.Filter := FRConst_OpenDiatogFilter;
  SaveDialog.Title  := FRConst_SaveDiatogTitle;
  ResizePaint := False;
  LastZoom     := 100;
  LastZoomType := ztNormal;
  FindText     := False;
  TextFound := False;
  {$IFNDEF Delphi2}
    OpenBtn.Flat   := True;
    SaveBtn.Flat   := True;
    PrintBtn.Flat  := True;
    FindBtn.Flat   := True;
    CloseBtn.Flat  := True;
    ScreenBtn.Flat := True;
  {$ENDIF}
  {$IFDEF Delphi4}
    OnMouseWheelUp   := MouseWheelUp;
    OnMouseWheelDown := MouseWheelDown;
    PopupMenu.Images := ImageList;
    {$IFDEF Delphi5}
      PopupMenu.AutoHotkeys := maManual;
    {$ENDIF}
  {$ENDIF}
  AddMenu (PopupMenu.Items,OpenBtn);
  AddMenu (PopupMenu.Items,SaveBtn);
  AddMenu (PopupMenu.Items,PrintBtn);
  AddMenu (PopupMenu.Items,FindBtn);
  P := TMenuItem.Create (Nil);
  P.Caption := '-';
  PopupMenu.Items.Add (P);
  For I := 0 To ZoomBtn.Items.Count-1 Do
    Begin
      P := TMenuItem.Create (Nil);
      P.Caption := ZoomBtn.Items[I];
      P.OnClick := MenuClick;
      PopupMenu.Items.Add (P);
    End;
  Panel := TPaintPanel.Create (Self);
  Panel.Parent      := Self;
  Panel.Align       := alClient;
  Panel.BevelOuter  := bvNone;
  Panel.Color       := clSilver;
  Panel.FullRepaint := False;
  Panel.NoPaint     := False;
  Panel.Drag        := False;
  Panel. NoSetY     := False;
  Panel.Cursor      := crMyCursor;
  PagePanel.BringToFront;
end;

procedure TStandardPreview.VScrollChange(Sender: TObject);
begin
  ScrollWindow (Panel.Handle,0,VPosition-VScroll.Position,Nil,Nil);
  VPosition := VScroll.Position;
  CheckZoom;
end;

procedure TStandardPreview.FormKeyDown(Sender: TObject; var Key: Word;
                                       Shift: TShiftState);
begin
  If Screen.ActiveControl<>ZoomBtn Then
    Case Key of
      VK_Home  : First;
      VK_End   : Last;
      VK_Up    : VScroll.Position := VScroll.Position - VScroll.SmallChange;
      VK_Down  : VScroll.Position := VScroll.Position + VScroll.SmallChange;
      VK_Prior : If ssCtrl in Shift Then Prior
                                    Else
                   VScroll.Position := VScroll.Position - VScroll.LargeChange;
      VK_Next  : If ssCtrl in Shift Then Next
                                    Else
                   VScroll.Position := VScroll.Position + VScroll.LargeChange;
      VK_Left  : HScroll.Position := HScroll.Position - HScroll.SmallChange;
      VK_Right : HScroll.Position := HScroll.Position + HScroll.SmallChange;
      VK_Escape: If EnableBigScreen Then EnableBigScreen := False
                                    Else Close;
      VK_F2    : Print;
      VK_F3    : Find;
      VK_F10   : Open;
      VK_F11   : Save;
      VK_SPACE : EnableBigScreen := Not EnableBigScreen;
      187      : Next;   // + key
      189      : Prior;  // - key
    End;
end;

procedure TStandardPreview.HScrollChange(Sender: TObject);
begin
  ScrollWindow (Panel.Handle,HPosition-HScroll.Position,0,Nil,Nil);
  HPosition := HScroll.Position;
  CheckZoom;
end;

procedure TStandardPreview.CloseBtnClick(Sender: TObject);
begin
  Close;
end;

procedure TStandardPreview.FormShow(Sender: TObject);
Var I, Dx, Dy: Integer;
begin
  FEnableBigScreen  := False;
  PagePanel.Visible := False;
  PagePanel.Top := 0;
  VPanel.Visible := True;
  HPanel.Visible := True;
  MaxWidth  := 0;
  MaxHeight := 0;
  For I := 0 To Report.EMFPages.Count-1 Do
    Begin
      Dx := Round (Report.EMFPages.Pages[I].pgWidth * mmToPixel);
      Dy := Round (Report.EMFPages.Pages[I].pgHeight * mmToPixel);
      If Dx>MaxWidth  Then MaxWidth  := Dx;
      If Dy>MaxHeight Then MaxHeight := Dy;
    End;
  VScroll.Position := 0;
  HScroll.Position := 0;
  VPosition := 0;
  HPosition := 0;
  HPanel.Height       := GetSystemMetrics (SM_CYHSCROLL);
  VPanel.Width        := GetSystemMetrics (SM_CXVSCROLL);
  ButtonsPanel.Height := HPanel.Height * 2;
  PrevBtn.Width       := ButtonsPanel.Width;
  PrevBtn.Height      := HPanel.Height;
  NextBtn.Width       := ButtonsPanel.Width;
  NextBtn.Top         := PrevBtn.Top + PrevBtn.Height;
  NextBtn.Height      := HPanel.Height;
  RightPanel.Width    := VPanel.Width;
  PageNo    := 1;
  FZoomType := ztNormal;
  Zoom      := 100;
end;

procedure TStandardPreview.PrevBtnClick(Sender: TObject);
begin
  Prior;
end;

procedure TStandardPreview.NextBtnClick(Sender: TObject);
begin
  Next;
end;

procedure TStandardPreview.ZoomBtnClick(Sender: TObject);
begin
  CheckZoom;
end;

procedure TStandardPreview.FormKeyPress(Sender: TObject; var Key: Char);
begin
  If Key=#13 Then Key := #0;
end;

procedure TStandardPreview.OpenBtnClick(Sender: TObject);
begin
  Open;
end;

procedure TStandardPreview.SaveBtnClick(Sender: TObject);
begin
  Save;
end;

procedure TStandardPreview.PrintBtnClick(Sender: TObject);
begin
  Print;
end;

procedure TStandardPreview.FormResize(Sender: TObject);
begin
  ResizePaint := True; //ZoomType in [ztOnePage,ztTwoPages];
  SetZoomFactor;
  If PagePanel.Visible Then PagePanel.Top := ClientHeight-PagePanel.Height;
end;

procedure TStandardPreview.ZoomBtnKeyPress(Sender: TObject; var Key: Char);
begin
  If Not (Key in ['0'..'9','%',DecimalSeparator,#8]) Then Key := #0;
end;

procedure TStandardPreview.FindBtnClick(Sender: TObject);
begin
  Find;
end;

procedure TStandardPreview.TimerTimer(Sender: TObject);
Var P: PRect;
    T: TPoint;
begin
  Inc (FoundTimer);
  New (P);
  CopyRect (P^,ExtRect);
  T := FindXY (PageFound);
  OffsetRect (P^,T.X,T.Y);
  InvalidateRect (Panel.Handle,P,False);
  Dispose (P);
  If FoundTimer=9 Then Begin
                         Timer.Enabled := False;
                         TextFound     := False;
                         InvalidateRect (Panel.Handle,P,False);
                       End;
end;

procedure TStandardPreview.PaintPagePaint(Sender: TObject);
Var S: String;
    R: TRect;
begin
  S := FRConst_Page + ' ' + IntToStr(PageNo) + '/' +
       IntToStr(Report.EMFPages.Count) + '          ';
  R := Rect(0,0,PaintPage.Width,PaintPage.Height);
  DrawText (PaintPage.Canvas.Handle,PChar(S),Length(S),R,DT_VCENTER);
  DrawText (PaintPage2.Canvas.Handle,PChar(S),Length(S),R,DT_VCENTER);
end;

procedure TStandardPreview.ScreenBtnClick(Sender: TObject);
begin
  EnableBigScreen := Not EnableBigScreen;
end;
// Protected procedures...
procedure TStandardPreview.SetBigScreen (Value:Boolean);
begin
  If FEnableBigScreen<>Value Then
    Begin
      If Value=True Then Begin
                           OldWindowState      := WindowState;
                           FOldEnableStatusBar := EnableStatusBar;
                           EnableStatusBar     := False;
                           WindowState         := wsMaximized;
                         End
                    Else EnableStatusBar     := FOldEnableStatusBar;
      If Not Value Then Begin
                          PagePanel.Visible := False;
                          PagePanel.Top := 0;
                          VPanel.Visible := True;
                          HPanel.Visible := True;
                          LeftPanel.Visible := FEnablePagePanel;
                          BorderStyle := bsSizeable;
                          WindowState := OldWindowState;
                        End
                   Else Begin
                          BorderStyle := bsNone;
                          VPanel.Visible := False;
                          HPanel.Visible := False;
                          If FEnablePagePanel Then
                            Begin
                              PagePanel.Top := ClientHeight-PagePanel.Height;
                              PagePanel.Visible := True;
                            End;
                        End;
    End;
  If (HScroll.Position<>0) Or (HScroll.Max<=70) Then Begin
                                                       Panel.NoSetY := True;
                                                       LastZoom:=0;
                                                       SetZoomFactor;
                                                       Panel.NoSetY := False;
                                                     End;
  FEnableBigScreen := Value;
end;

procedure TStandardPreview.SetStatusBar (Value:Boolean);
begin
  If Value Then StatusBar.Height := 32
           Else StatusBar.Height := 0;
end;

procedure TStandardPreview.SetPagePanel (Value:Boolean);
begin
  If Value<>FEnablePagePanel Then
    Begin
      If Value Then Begin
                      If Not FEnableBigScreen Then LeftPanel.Visible := True
                                              Else
                        Begin
                          PagePanel.Top := ClientHeight-PagePanel.Height;
                          PagePanel.Visible := True;
                        End;
                    End
               Else Begin
                      PagePanel.Visible := False;
                      PagePanel.Top := 0;
                      LeftPanel.Visible := False;
                    End;
      FEnablePagePanel := Value;
    End;
end;

procedure TStandardPreview.SetSmallChange (Value:Integer);
begin
  If Value<>FSmallChange Then
    Begin
      FSmallChange := Value;
      HScroll.SmallChange := Value;
      VScroll.SmallChange := Value;
    End;
end;

procedure TStandardPreview.SetLargeChange (Value:Integer);
begin
  If Value<>FLargeChange Then
    Begin
      FLargeChange := Value;
      HScroll.LargeChange := Value;
      VScroll.LargeChange := Value;
    End;
end;

// Public procedures...
procedure TStandardPreview.Open;
Var OldDir: String;
begin
  CheckZoom;
  OldDir := GetCurrentDir;
  With OpenDialog Do
    Begin
      If Execute Then
        Begin
          Application.ProcessMessages;
          Report.LoadPreparedReport (FileName);
          LastZoom := 0;
          FormShow(Nil);
        End;
    End;
  SetCurrentDir (OldDir);
end;

procedure TStandardPreview.Save;
Var I: Integer;
    OldDir: String;
begin
  CheckZoom;
  OldDir := GetCurrentDir;
  With SaveDialog Do
    Begin
      Filter := FRConst_SaveDiatogFilter;
      For I := 0 To frFiltersCount-1 Do
        Filter := Filter + '|' + frFilters[I].FilterDesc + '|' + frFilters[I].FilterExt;
      FilterIndex := 1;
      If Execute Then
        Begin
          Application.ProcessMessages;
          If FilterIndex = 1 Then Report.SavePreparedReport (FileName)
                             Else
          If FilterIndex = 2 Then
            Begin
              If Pos('.EMF',UpperCase(FileName))<>Length(FileName)-3 Then
                Report.EMFPages[PageNo-1].Pict.SaveToFile (FileName + '.emf')
                                                                     Else
                Report.EMFPages[PageNo-1].Pict.SaveToFile (FileName)
            End
                             Else
            Begin
              FilterIndex := FilterIndex - 1;
              Report.ExportTo(frFilters[FilterIndex-2].ClassRef,
                ChangeFileExt(FileName,Copy(frFilters[FilterIndex-2].FilterExt,2,255)));
            End;
        End;
    End;
  SetCurrentDir (OldDir);
end;

procedure TStandardPreview.Print;
Var Copy: Integer;
begin
  CheckZoom;
  With PrintDialog Do
    Begin
      MinPage  := 1;
      MaxPage  := Report.EMFPages.Count;
      FromPage := MinPage;
      ToPage   := MaxPage;
      Copies   := 1;
      If Execute Then
        Begin
          Application.ProcessMessages;
          Copy   := Copies;
          Copies := 1;
          Report.PrintPreparedReport (FromPage,ToPage,Copy,Collate);
        End;
    End;
end;

procedure TStandardPreview.Find;
Var I: Integer;
    OldText: String;
    OldCase, OldSub, OldNext, OldPrev: Boolean;
    OldDir: Integer;
    FindForm: TFindForm;
begin
  FindForm := TFindForm.Create (Nil);
  With FindForm Do
    Begin
      FindCase.Checked := FindData.OFindCase;
      FindSub.Checked  := FindData.OFindSub;
      FindFirstPage.Checked   := False;
      FindCurrentPage.Checked := False;
      FindNext.Checked        := False;
      FindPrevious.Checked    := False;
      Case FindData.OFindOrigin Of
        0: FindFirstPage.Checked   := True;
        1: FindCurrentPage.Checked := True;
        2: FindNext.Checked        := True;
        3: FindPrevious.Checked    := True;
      End;
      FindNext.Enabled := FindData.EnbNext;
      FindPrevious.Enabled := FindData.EnbPrev;
      FindText.Text := FindData.OText;
      FindText.Items.Assign (FindData.OFindText);
      OldText := FindText.Text;
      OldCase := FindCase.Checked;
      OldSub  := FindSub.Checked;
      OldNext := FindNext.Enabled;
      OldPrev := FindPrevious.Enabled;
      If FindFirstPage.Checked Then OldDir := 0
                               Else
      If FindCurrentPage.Checked Then OldDir := 1
                                 Else
      If FindNext.Checked Then OldDir := 2
                          Else OldDir := 3;
    End;
  If FindForm.ShowModal=mrOK Then
    Begin
      If TextFound Then
        Begin
          FoundTimer := 8;
          TimerTimer (Nil);
        End;
      Application.ProcessMessages;
      FindText  := True;
      TextFound := False;
      TextToFind    := FindForm.FindText.Text;
      CaseSensitive := FindForm.FindCase.Checked;
      SubString     := FindForm.FindSub.Checked;
      NextItem      := FindForm.FindNext.Checked;
      PrevItem      := FindForm.FindPrevious.Checked;
      CurrPage      := FindForm.FindCurrentPage.Checked;
      CurrObject    := 0;
      I := 0;
      While (I<=Report.EMFPages.Count-1) And (Not TextFound) Do
        Begin
          DrawPage (Panel.Canvas.Handle,I,Rect(0,0,10,10));
          Inc (I);
        End;
      FindText := False;
      If TextFound Then
        Begin
          FindForm.FindNext.Enabled     := True;
          FindForm.FindPrevious.Enabled := (FindForm.FindNext.Checked) Or
                                           (FindForm.FindPrevious.Checked);
          If FindForm.FindText.Items.IndexOf (FindForm.FindText.Text)=-1 Then
            FindForm.FindText.Items.Add (FindForm.FindText.Text);
          If PageNo<>PageFound+1 Then SetPageNumber (PageFound+1);
          HScroll.Position := HScroll.Position + FindXY(PageFound).X + ExtRect.Left -
                              (Panel.Width-(ExtRect.Right - ExtRect.Left)) Div 2;
          VScroll.Position := VScroll.Position + FindXY(PageFound).Y + ExtRect.Top -
                             (Panel.Height-(ExtRect.Bottom - ExtRect.Top)) Div 2;
          TimerTimer (Nil);
        End
                   Else
      Application.MessageBox (PChar(FRConst_FindError),PChar(FRConst_FindTitle),
                  MB_OK or MB_ICONWARNING Or MB_SYSTEMMODAL);
    End
                             Else
    With FindForm Do
      Begin
        FindText.Text        := OldText;
        FindCase.Checked     := OldCase;
        FindSub.Checked      := OldSub;
        FindNext.Enabled     := OldNext;
        FindPrevious.Enabled := OldPrev;
        Case OldDir of
          0: FindFirstPage.Checked   := True;
          1: FindCurrentPage.Checked := True;
          2: FindNext.Checked        := True;
          3: FindPrevious.Checked    := True;
        End;
      End;
  With FindForm Do
    Begin
      FindData.OFindCase := FindCase.Checked;
      FindData.OFindSub  := FindSub.Checked;
      If FindFirstPage.Checked Then FindData.OFindOrigin := 0
                               Else
      If FindCurrentPage.Checked Then FindData.OFindOrigin := 1
                                 Else
      If FindNext.Checked Then FindData.OFindOrigin := 2
                          Else FindData.OFindOrigin := 3;
      FindData.EnbNext := FindNext.Enabled;
      FindData.EnbPrev := FindPrevious.Enabled;
      FindData.OText   := FindText.Text;
      FindData.OFindText.Assign (FindText.Items);
    End;
  FindForm.Free;
end;

procedure TStandardPreview.First;
begin
  SetPageNumber (1);
end;

procedure TStandardPreview.Last;
begin
  SetPageNumber (Report.EMFPages.Count);
end;

procedure TStandardPreview.Next;
begin
  If PageNo<Report.EMFPages.Count Then
    If ZoomType<>ztTwoPages Then SetPageNumber (PageNo+1)
                            Else SetPageNumber (PageNo+2);
  CheckZoom;
end;

procedure TStandardPreview.Prior;
begin
  If PageNo>1 Then
    If ZoomType<>ztTwoPages Then SetPageNumber (PageNo-1)
                            Else SetPageNumber (PageNo-2);
  CheckZoom;
end;

procedure TStandardPreview.ZoomWidth;
begin
  ZoomBtn.Text := FRConst_ZoomWidth;
  FZoomType := ztWidth;
  SetZoomFactor;
end;

procedure TStandardPreview.ZoomOnePage;
begin
  ZoomBtn.Text := FRConst_ZoomOnePage;
  FZoomType := ztOnePage;
  SetZoomFactor;
end;

procedure TStandardPreview.ZoomTwoPages;
begin
  ZoomBtn.Text := FRConst_ZoomTwoPages;
  FZoomType := ztTwoPages;
  SetZoomFactor;
end;
//------------------------------------------------------------------------------
Function GetOSVersion:Cardinal;
Var Info: TOsVersionInfo;
Begin
  Info.dwOSVersionInfoSize := SizeOf(Info);
  GetVersionEx (Info);
  Result := Info.dwPlatformId;
End;

procedure TStandardPreview.ZoomBtnKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  If Key=VK_Return Then CheckZoom;
end;

procedure TStandardPreview.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  If Timer.Enabled Then Begin
                          FoundTimer := 8;
                          TimerTimer (Nil);
                        End;
end;



procedure TStandardPreview.HI_PanelResize(Sender: TObject);
begin
  With HScroll Do
    Begin
      Left := 0;
      Top  := 0;
      Width  := HI_Panel.Width;
      Height := HI_Panel.Height;
    End;
end;

procedure TStandardPreview.VI_PanelResize(Sender: TObject);
begin
  With VScroll Do
    Begin
      Left := 0;
      Top  := 0;
      Width  := VI_Panel.Width;
      Height := VI_Panel.Height;
    End;
end;

Initialization
  WinNT := GetOSVersion >= VER_PLATFORM_WIN32_NT;
  Screen.Cursors [crMyCursor] := LoadCursor (HInstance,'HAND');
  SetPreviewClass (TStandardPreview);

Finalization
  DestroyCursor (Screen.Cursors [crMyCursor]);

end.