{*******************************************
 *             FuzzyReport 2.0             *
 *                                         *
 *  Copyright (c) 2000 by Fabio Dell'Aria  *
 *                                         *
 *-----------------------------------------*
 * For to use this source code, you must   *
 * read and agree all license conditions.  *
 *******************************************}

unit QRtoFR;

{$I FR_Vers.inc}

{$R QRConverter.res}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Quickrpt, QrCtrls, DB, Printers, FR_Desgn, FR_Class,
  {$IFNDEF DELPHI6}DsgnIntf{$ELSE}DesignIntf, DesignEditors{$ENDIF},
  FR_Shape, FR_Rich, FR_Const;

type

  RepType = TQuickRep;

  TQRtoFRConverter = class(TComponent)
  private
    { Private declarations }
    FQuickReport: RepType;
    FReport: TfrReport;
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    procedure ShowReport;
  published
    { Published declarations }
  end;

procedure Register;

implementation

uses FR_SelQR;

type
  TEditor = class(TComponentEditor)
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

procedure TEditor.ExecuteVerb(Index: Integer);
var
  I, iRepID: Integer;
  sElencoReport, sElencoID: TStringList;
begin
  { 08/03/2001 - Zanella
    Cerca i report da convertire nel form su cui  stato posizionato il
    componente QrToFr }
  with TQRtoFRConverter(Component) do
    begin
      sElencoReport := TStringList.Create;
      sElencoID := TStringList.Create;

      try
        { Scorre i componenti presenti sul form che contiene il componente }
        for I := 0 to Pred(TForm(Owner).ControlCount) do
          begin
            { Se trova un QuickReport aggiunge a sElencoReport il suo nome e ad
              sElencoID il suo ID }
            if TForm(Owner).Controls[I] is RepType then
              begin
                sElencoReport.Add(RepType(TForm(Owner).Controls[I]).Name);
                sElencoID.Add(IntToStr(I));
              end;
          end;

        { Controlla quanti report sono stati trovati }
        case sElencoReport.Count of
          0: // 0 --> Nessun report, visualizza un errore
            begin
              Application.MessageBox(PChar('You must select a QuickReport'),
                PChar('Error'), MB_OK or MB_ICONERROR or MB_APPLMODAL);
            end;
          1: // 1 --> Un report, lo converte direttamente
            begin
              iRepID := StrToInt(sElencoID[0]);

              { Assegna a FQuickReport il report da convertire }
              FQuickReport := RepType(TForm(Owner).Controls[iRepID]);
              { Lancia la routine di conversione del report }
              ShowReport;
            end;
        else
          begin
            { Se i report da convertire sono pi di uno visualizza una maschera che
              permette di scegliere quali convertire }
            with TfrmSelReport.Create(nil) do
              begin
                try
                  { Visualizza l'elenco dei report }
                  Elenco := sElencoReport;
                  if ShowModal = mrOK then
                    begin
                      { Aggiorna l'elenco dei report in modo da determinare quali
                        convertire }
                      AggiornaElenco(sElencoReport, sElencoID);
                    end
                  else
                    begin
                      sElencoID.Clear;
                    end;
                finally
                  Release;
                end;
              end;

            { Converte tutti i report selezionati }
            for I := 0 to Pred(sElencoID.Count) do
              begin
                iRepID := StrToInt(sElencoID[I]);

                { Assegna a FQuickReport il report da convertire }
                FQuickReport := RepType(TForm(Owner).Controls[iRepID]);
                { Lancia la routine di conversione del report }
                ShowReport;
              end;
          end;
        end;
      finally
        sElencoReport.Free;
        sElencoID.Free;
      end;
    end;
  { 08/03/2001 - Fine }
end;

function TEditor.GetVerb(Index: Integer): string;
begin
  Result := FRConst_ReportConvert;
end;

function TEditor.GetVerbCount: Integer;
begin
  Result := 1;
end;

constructor TQRtoFRConverter.Create(AOwner: TComponent);
begin
  inherited;
  if not (AOwner is TForm) then
    begin
      Application.MessageBox(PChar('Put this component into the Form with QuickReports.'),
        PChar('Warning'), MB_OK or MB_ICONWARNING or MB_APPLMODAL);
      FQuickReport := nil;
    end
end;

procedure TQRtoFRConverter.ShowReport;
const
  Converti = 794 / 210;
  Bande: array[0..9] of TfrBandType = (btReportTitle, btPageHeader, btMasterData,
    btPageFooter, btReportSummary, btGroupHeader,
    btGroupFooter, btDetailData, btStaticColumnHeader,
    btOverlay);
var
  I, J, L, YBand{$IFDEF Delphi4}, H{$ENDIF}: Integer;
  Memo, BackBand: TfrMemoView;
  Band: TfrBandView;
  S: string;
  List: TStringList;
  OldZoom: Integer;
  OldUnits: TQRUnit;
  OKAutoStretch: Boolean;
  DSet: TfrDBDataSet;
  Error: string;

  {$HINTS OFF}
  function ConvertiEspressione(S: string): string;
  const
    ConvFunc: array[0..1, 0..14] of string = (('AVERAGE', 'COPY', 'COUNT', 'FORMATNUMERIC',
      'FRAC', 'IF', 'INT', 'LOWER', 'MAX', 'MIN',
      'PRETTY', 'SQRT', 'STR', 'SUM', 'UPPER'),
      ('AVG', 'COPY', 'COUNT', 'FORMATFLOAT',
      'FRAC', 'IF', 'INT', 'LOWERCASE', 'MAX',
      'MIN', 'NAMECASE', 'SQRT', 'STR', 'SUM',
      'UPPERCASE'));
  var
    J, C, P: Integer;

    function NewPos(Sub, Big: string; N: Integer): Integer;
    var
      T: Integer;
    begin
      T := Pos(Sub, Copy(Big, N, Length(Big) - N + 1));
      if T <> 0 then
        Result := (N - 1) + T
      else
        Result := 0;
    end;

    function InQuote(S: string; N: Integer): Boolean;
    var
      G, Q, W: Integer;
    begin
      Q := 0;
      W := 0;
      for G := 1 to N - 1 do
        if S[G] = '''' then
          Inc(Q)
        else
          if S[G] = '"' then Inc(W);
      Result := (Q mod 2 = 1) or (W mod 2 = 1);
    end;

  begin
    if Trim(S) = '' then
      Result := ''
    else
      if Uppercase(TRIM(S)) = 'COUNT' then
        Result := '[COUNT()]'
      else
        if Pos('(', S) = 0 then
          Result := '[' + S + ']'
        else
          begin
            for J := 0 to 14 do
              begin
                C := 1;
                repeat
                  P := NewPos(ConvFunc[0, J], UpperCase(S), C);
                  if P <> 0 then
                    if not InQuote(S, P) then
                      begin
                        S := Copy(S, 1, P - 1) + ConvFunc[1, J] +
                          Copy(S, P + Length(ConvFunc[0, J]), Length(S));
                        C := P + Length(ConvFunc[1, J]);
                      end
                    else
                      C := P + Length(ConvFunc[0, J]);
                until P = 0;
              end;
            Result := '[' + S + ']';
          end;
  end;
  {$HINTS ON}

begin
  Error := '';
  if FQuickReport = nil then
    begin
      Application.MessageBox(FRConst_ConvLimit, FRConst_Attenction,
        MB_OK or MB_ICONWARNING);
      Exit;
    end;
  FReport := TfrReport.Create(TControl(Self));
  FReport.Name := 'Report_01';
  FReport.DoublePass := False;
  with FReport.Pages.Pages[0] do
    begin
      OldZoom := FQuickReport.Zoom;
      OldUnits := FQuickReport.Units;
      FQuickReport.Units := mm;
      FQuickReport.Zoom := 100;
      PHonFirst := (FirstPageHeader in FQuickReport.Options);
      PFonLast := (LastPageFooter in FQuickReport.Options);
      ChangePaper($100, Round(FQuickReport.Page.Width * 10),
        Round(FQuickReport.Page.Length * 10),
        FQuickReport.Page.Orientation);
      UseMargins := True;
      PHonFirst := FirstPageHeader in FQuickReport.Options;
      PFonLast := LastPageFooter in FQuickReport.Options;
      if FQuickReport.Page.Orientation = poLandscape then
        begin
          pgMargins.Left := Round(FQuickReport.Page.BottomMargin * Converti);
          pgMargins.Top := Round(FQuickReport.Page.LeftMargin * Converti);
          pgMargins.Right := Round(FQuickReport.Page.TopMargin * Converti);
          pgMargins.Bottom := Round(FQuickReport.Page.RightMargin * Converti)
        end
      else
        begin
          pgMargins.Left := Round(FQuickReport.Page.LeftMargin * Converti);
          pgMargins.Top := Round(FQuickReport.Page.TopMargin * Converti);
          pgMargins.Right := Round(FQuickReport.Page.RightMargin * Converti);
          pgMargins.Bottom := Round(FQuickReport.Page.BottomMargin * Converti)
        end;
      List := TStringList.Create;
      List.Sorted := True;
      try
        for I := 0 to FQuickReport.ControlCount - 1 do
          begin
            if (FQuickReport.Controls[I] is TQrCustomBand) and
              (Integer(TQrCustomBand(FQuickReport.Controls[I]).BandType) < 10) then
              begin
                S := IntToStr(TQrCustomBand(FQuickReport.Controls[I]).Top);
                for J := Length(S) + 1 to 5 do
                  S := '0' + S;
                List.AddObject(S, FQuickReport.Controls[I])
              end
          end;
        YBand := 19;
        for I := 0 to List.Count - 1 do
          begin
            OKAutoStretch := False;
            Objects.Add(frCreateObject(gtBand, ''));
            Band := Objects.Last;
            Band.ForceNewPage := TQrCustomBand(List.Objects[I]).ForceNewPage;
            Band.BandType := Bande[Integer(TQrCustomBand(List.Objects[I]).BandType)];
            Band.Left := TQrCustomBand(List.Objects[I]).Left;
            Band.Width := TQrCustomBand(List.Objects[I]).Width;
            Band.Top := YBand;
            Band.Height := TQrCustomBand(List.Objects[I]).Height;
            Band.Name := TQrCustomBand(List.Objects[I]).Name;
            Inc(YBand, Band.Height + 24);
            Objects.Add(frCreateObject(gtMemo, ''));
            BackBand := Objects.Last;
            BackBand.Left := Band.Left;
            BackBand.Width := Band.Width;
            BackBand.Top := Band.Top;
            BackBand.Height := Band.Height;
            BackBand.Color := TQrCustomBand(List.Objects[I]).Color;
            BackBand.FrameWidth := TQrCustomBand(List.Objects[I]).Frame.Width;
            BackBand.FrameColor := TQrCustomBand(List.Objects[I]).Frame.Color;
            BackBand.DrawFrameRight := TQrCustomBand(List.Objects[I]).Frame.DrawRight;
            BackBand.DrawFrameLeft := TQrCustomBand(List.Objects[I]).Frame.DrawLeft;
            BackBand.DrawFrameTop := TQrCustomBand(List.Objects[I]).Frame.DrawTop;
            BackBand.DrawFrameBottom := TQrCustomBand(List.Objects[I]).Frame.DrawBottom;
            if Integer(TQrCustomBand(List.Objects[I]).BandType) = 2 then
              begin
                DSet := TfrDBDataSet.Create(nil);
                DSet.DataSet := FQuickReport.DataSet;
                FReport.SetDataSetBand(Band.Name, DSet)
              end
            else
              if Integer(TQrCustomBand(List.Objects[I]).BandType) = 5 then
                begin
                  Band.FormatStr := ConvertiEspressione(TQrGroup(List.Objects[I]).Expression);
                  if Band.FormatStr <> '' then
                    if Band.FormatStr[1] = '[' then
                      Band.FormatStr := Copy(Band.FormatStr, 2, Length(Band.FormatStr) - 2);
                end
              else
                if Integer(TQrCustomBand(List.Objects[I]).BandType) = 7 then
                  begin
                    DSet := TfrDBDataSet.Create(nil);
                    DSet.DataSet := TQrSubDetail(List.Objects[I]).DataSet;
                    FReport.SetDataSetBand(Band.Name, DSet);
                  end;
            for J := 0 to TQrCustomBand(List.Objects[I]).ControlCount - 1 do
              begin
                if (TQrCustomBand(List.Objects[I]).Controls[J] is TQRPrintable) then
                  begin
                    if (TQrCustomBand(List.Objects[I]).Controls[J] is TQrImage) or
                      (TQrCustomBand(List.Objects[I]).Controls[J] is TQrDBImage) then
                      Objects.Add(frCreateObject(gtPicture, ''))
                    else
                      if (TQrCustomBand(List.Objects[I]).Controls[J] is TQRDBRichText) or
                        (TQrCustomBand(List.Objects[I]).Controls[J] is TQRRichText) then
                        Objects.Add(frCreateObject(gtAddIn, 'TfrRichView'))
                      else
                        if (TQrCustomBand(List.Objects[I]).Controls[J] is TQRShape) and
                          (TQRShape(TQrCustomBand(List.Objects[I]).Controls[J]).Shape <> qrsRightAndLeft) and
                          (TQRShape(TQrCustomBand(List.Objects[I]).Controls[J]).Shape <> qrsTopAndBottom) then
                          Objects.Add(frCreateObject(gtAddIn, 'TfrShapeView'))
                        else
                          Objects.Add(frCreateObject(gtMemo, ''));
                    Memo := Objects.Last;
                    Memo.Name := TQRPrintable(TQrCustomBand(List.Objects[I]).Controls[J]).Name;
                    Memo.Top := TQRPrintable(TQrCustomBand(List.Objects[I]).Controls[J]).Top + Band.Top;
                    Memo.Height := TQRPrintable(TQrCustomBand(List.Objects[I]).Controls[J]).Height;
                    Memo.Left := TQRPrintable(TQrCustomBand(List.Objects[I]).Controls[J]).Left + Band.Left;
                    Memo.Width := TQRPrintable(TQrCustomBand(List.Objects[I]).Controls[J]).Width;
                    if Memo.Top < Band.Top then Memo.Top := Band.Top;
                    Memo.FrameWidth := TQRPrintable(TQrCustomBand(List.Objects[I]).Controls[J]).Frame.Width;
                    Memo.FrameColor := TQRPrintable(TQrCustomBand(List.Objects[I]).Controls[J]).Frame.Color;
                    Memo.FrameStyle := TQRPrintable(TQrCustomBand(List.Objects[I]).Controls[J]).Frame.Style;
                    Memo.Visible := TQRPrintable(TQrCustomBand(List.Objects[I]).Controls[J]).Enabled;
                    Memo.DrawFrameRight :=
                      TQRPrintable(TQrCustomBand(List.Objects[I]).Controls[J]).Frame.DrawRight;
                    Memo.DrawFrameLeft :=
                      TQRPrintable(TQrCustomBand(List.Objects[I]).Controls[J]).Frame.DrawLeft;
                    Memo.DrawFrameTop :=
                      TQRPrintable(TQrCustomBand(List.Objects[I]).Controls[J]).Frame.DrawTop;
                    Memo.DrawFrameBottom :=
                      TQRPrintable(TQrCustomBand(List.Objects[I]).Controls[J]).Frame.DrawBottom;
                    if (TQrCustomBand(List.Objects[I]).Controls[J] is TQRCustomLabel) then
                      begin
                        Memo.Color := TQRCustomLabel(TQrCustomBand(List.Objects[I]).Controls[J]).Color;
                        Memo.Font.Name := TQRCustomLabel(TQrCustomBand(List.Objects[I]).Controls[J]).Font.Name;
                        Memo.Font.Color := TQRCustomLabel(TQrCustomBand(List.Objects[I]).Controls[J]).Font.Color;
                        Memo.Font.Size := TQRCustomLabel(TQrCustomBand(List.Objects[I]).Controls[J]).Font.Size;
                        Memo.Font.Style := TQRCustomLabel(TQrCustomBand(List.Objects[I]).Controls[J]).Font.Style;
                        Memo.AutoStretch := TQRCustomLabel(TQrCustomBand(List.Objects[I]).Controls[J]).AutoStretch;
                        if Memo.AutoStretch then OKAutoStretch := True;
                        Memo.AutoSize := TQRCustomLabel(TQrCustomBand(List.Objects[I]).Controls[J]).AutoSize;
                        Memo.WordWrap := TQRCustomLabel(TQrCustomBand(List.Objects[I]).Controls[J]).WordWrap;
                        case TQRCustomLabel(TQrCustomBand(List.Objects[I]).Controls[J]).Alignment of
                          TAlignment(taLeftJustify): Memo.Alignment := TFrAlignment(taLeftJustify);
                          TAlignment(taRightJustify): Memo.Alignment := TFrAlignment(taRightJustify);
                          TAlignment(taCenter): Memo.Alignment := TFrAlignment(taCenter);
                        end;
                      end;
                    if (TQrCustomBand(List.Objects[I]).Controls[J] is TQRRichText) then
                      begin
                        if TQRRichText(TQrCustomBand(List.Objects[I]).
                          Controls[J]).ParentRichEdit <> nil then
                          {$IFDEF RX_RICH}
                          AssignRichToRxRich(TQRRichText(TQrCustomBand(List.Objects[I]).
                            Controls[J]).ParentRichEdit,
                            TfrRichView(Memo).RichEdit)
                            {$ELSE}
                          AssignRich(TfrRichView(Memo).RichEdit,
                            TQRRichText(TQrCustomBand(List.Objects[I]).
                            Controls[J]).ParentRichEdit)
                            {$ENDIF}
                        else
                          TfrRichView(Memo).RichEdit.Lines.Assign
                            (TQRRichText(TQrCustomBand(List.Objects[I]).Controls[J]).Lines);
                        TfrRichView(Memo).CalcExpr := False;
                      end;
                    if (TQrCustomBand(List.Objects[I]).Controls[J] is TQRDBRichText) then
                      begin
                        TfrRichView(Memo).RichEdit.Lines.Add(
                          '["' + TQRDBRichText(TQrCustomBand(List.Objects[I]).Controls[J]).DataField + '"]');
                      end;
                    {$IFDEF Delphi4}
                    if (TQrCustomBand(List.Objects[I]).Controls[J] is TQRExprMemo) then
                      begin
                        for H := 0 to TQRExprMemo(TQrCustomBand(List.Objects[I]).Controls[J]).Lines.Count - 1 do
                          begin
                            S := TQRExprMemo(TQrCustomBand(List.Objects[I]).Controls[J]).Lines[H];
                            repeat
                              if (Pos('{', S) > 0) and (Pos('}', S) > 0) then
                                begin
                                  S := Copy(S, 1, Pos('{', S) - 1) + '[' +
                                    ConvertiEspressione(Copy(S, Pos('{', S) + 1,
                                    Pos('}', S) - Pos('{', S))) +
                                    ']' + Copy(S, Pos('}', S) + 1, Length(S));
                                end;
                            until (Pos('}', S) = 0) or (Pos('{', S) = 0);
                            Memo.Memo.Add(S);
                          end;
                      end;
                    {$ENDIF}
                    if (TQrCustomBand(List.Objects[I]).Controls[J] is TQRShape) then
                      begin
                        Memo.Color := TQRShape(TQrCustomBand(List.Objects[I]).Controls[J]).Brush.Color;
                        Memo.FrameColor := TQRShape(TQrCustomBand(List.Objects[I]).Controls[J]).Pen.Color;
                        Memo.FrameWidth := TQRShape(TQrCustomBand(List.Objects[I]).Controls[J]).Pen.Width;
                        TfrShapeView(Memo).Trans := False;
                        if (TQRShape(TQrCustomBand(List.Objects[I]).Controls[J]).Shape <> qrsRightAndLeft) and
                          (TQRShape(TQrCustomBand(List.Objects[I]).Controls[J]).Shape <> qrsTopAndBottom) then
                          begin
                            case TQRShape(TQrCustomBand(List.Objects[I]).Controls[J]).Shape of
                              qrsCircle: TfrShapeView(Memo).TypeShape := shCircle;
                              qrsRectangle: TfrShapeView(Memo).TypeShape := shRectangle;
                              qrsVertLine:
                                begin
                                  Memo.Left := Memo.Left + Memo.Width div 2;
                                  Memo.Width := 1;
                                end;
                              qrsHorLine:
                                begin
                                  Memo.Top := Memo.Top + Memo.Height div 2;
                                  Memo.Height := 1;
                                end;
                            end;
                          end
                        else
                          begin
                            case TQRShape(TQrCustomBand(List.Objects[I]).Controls[J]).Shape of
                              qrsRightAndLeft:
                                begin
                                  Memo.DrawFrameRight := True;
                                  Memo.DrawFrameLeft := True;
                                end;
                              qrsTopAndBottom:
                                begin
                                  Memo.DrawFrameTop := True;
                                  Memo.DrawFrameBottom := True;
                                end;
                            end;
                          end
                      end
                    else
                      if (TQrCustomBand(List.Objects[I]).Controls[J] is TQrImage) then
                        begin
                          TfrPictureView(Memo).Picture.Assign(TQrImage(TQrCustomBand(List.Objects[I]).Controls[J]).Picture);
                          TfrPictureView(Memo).Center := TQrImage(TQrCustomBand(List.Objects[I]).Controls[J]).Center;
                          TfrPictureView(Memo).Stretch := TQrImage(TQrCustomBand(List.Objects[I]).Controls[J]).Stretch;
                        end
                      else
                        if (TQrCustomBand(List.Objects[I]).Controls[J] is TQrDBImage) then
                          begin
                            Memo.Caption := '["' + TQrDBImage(TQrCustomBand(List.Objects[I]).Controls[J]).DataField + '"]';
                            TfrPictureView(Memo).Center := TQrImage(TQrCustomBand(List.Objects[I]).Controls[J]).Center;
                            TfrPictureView(Memo).Stretch := TQrImage(TQrCustomBand(List.Objects[I]).Controls[J]).Stretch;
                          end
                        else

                          if (TQrCustomBand(List.Objects[I]).Controls[J] is TQrLabel) then
                            begin
                              Memo.Caption := TQrLabel(TQrCustomBand(List.Objects[I]).Controls[J]).Caption;
                              if TQRLabel(TQrCustomBand(List.Objects[I]).Controls[J]).Transparent then
                                Memo.Color := clNone;
                            end
                          else
                            if (TQrCustomBand(List.Objects[I]).Controls[J] is TQRDBText) then
                              begin
                                if TQRDBText(TQrCustomBand(List.Objects[I]).
                                  Controls[J]).DataSet <> nil then
                                  Memo.Caption := '[' + TQRDBText(TQrCustomBand(List.Objects[I]).Controls[J]).DataSet.Owner.Name + '.' +
                                    TQRDBText(TQrCustomBand(List.Objects[I]).Controls[J]).DataSet.Name + '.' +
                                    '"' + TQRDBText(TQrCustomBand(List.Objects[I]).Controls[J]).DataField + '"]'
                                else
                                  begin
                                    Error := FRConst_ConvControlName +
                                      TQRDBText(TQrCustomBand(List.Objects[I]).
                                      Controls[J]).Name +
                                      FRConst_ConvNoDataSet;
                                    raise Exception.Create('');
                                  end;
                                Memo.Color := clNone;
                              end
                            else
                              if (TQrCustomBand(List.Objects[I]).Controls[J] is TQRExpr) then
                                begin
                                  Memo.Caption := ConvertiEspressione(TQRExpr(TQrCustomBand(List.Objects[I]).Controls[J]).Expression);
                                  if TQRExpr(TQrCustomBand(List.Objects[I]).Controls[J]).Transparent then
                                    Memo.Color := clNone;
                                end
                              else
                                if (TQrCustomBand(List.Objects[I]).Controls[J] is TQRMemo) then
                                  begin
                                    for L := 0 to TQRMemo(TQrCustomBand(List.Objects[I]).Controls[J]).Lines.Count - 1 do
                                      Memo.Memo.Add(TQRMemo(TQrCustomBand(List.Objects[I]).Controls[J]).Lines[L]);
                                    if TQRMemo(TQrCustomBand(List.Objects[I]).Controls[J]).Transparent then
                                      Memo.Color := clNone;
                                  end
                                else
                                  if (TQrCustomBand(List.Objects[I]).Controls[J] is TQRSysData) then
                                    begin
                                      S := TQRSysData(TQrCustomBand(List.Objects[I]).Controls[J]).Text;
                                      case TQRSysData(TQrCustomBand(List.Objects[I]).Controls[J]).Data of
                                        qrsDate: S := S + '[DATE]';
                                        qrsDateTime: S := S + '[DATE] [TIME]';
                                        qrsTime: S := S + '[TIME]';
                                        qrsPageNumber: S := S + '[PAGE#]';
                                        //                      qrsPageCount   : S:=S+'[TOTALPAGES]';
                                        qrsDetailNo: S := S + '[LINE#]';
                                        //                      qrsColumnNo    : S:=S+'[COLUMN#]';
                                        qrsDetailCount: S := S + '[LINETHROUGH#]';
                                      end;
                                      Memo.Caption := S;
                                      if TQRSysData(TQrCustomBand(List.Objects[I]).Controls[J]).Transparent then
                                        Memo.Color := clNone;
                                    end;
                  end
              end;
            Band.AutoStretch := OKAutoStretch;
            BackBand.AutoStretch := OKAutoStretch;
          end;
        if (FReport.ChekBandType(btDetailData, 0)) and
          (not FReport.ChekBandType(btMasterData, 0)) then
          for I := 0 to Objects.Count - 1 do
            if (TfrView(Objects[I]).Typ = gtBand) and
              (TfrBandView(Objects[I]).BandType = btDetailData) then
              TfrBandView(Objects[I]).BandType := btMasterData;
      except
        Application.MessageBox(PChar(Error), FRConst_Attenction,
          MB_OK or MB_ICONERROR);
      end;
    end;
  List.Free;
  FQuickReport.Zoom := OldZoom;
  FQuickReport.Units := OldUnits;
  if Error = '' then
    begin
      FReport.DoublePass := False;
      FReport.FileName := FQuickReport.Name + '.frf';
      FReport.DesignReport;
    end;
  FReport.Free;
end;

procedure Register;
begin
  RegisterComponents(FRConst_ReportName, [TQRtoFRConverter]);
  RegisterComponentEditor(TQRtoFRConverter, TEditor);
end;

end.

