
{*****************************************}
{                                         }
{             FastReport v2.2             }
{            Various routines             }
{                                         }
{  Copyright (c) 1998-99 by Tzyganenko A. }
{                                         }
{*****************************************}

unit FR_Utils;

interface

{$I FR_Vers.inc}

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls,
  DB, Forms, StdCtrls, ClipBrd, Menus{$IFNDEF Delphi2}, Buttons{$ENDIF};

procedure AddDescription(FunctionName, Category, Help: string; ParamNum: Integer);
procedure frReadMemo(Stream: TStream; l: TStrings);
procedure frWriteMemo(Stream: TStream; l: TStrings);
procedure frEnableControls(c: array of TControl; e: Boolean);
function frGetDataSet(ComplexName: string): TDataSet;
procedure frGetDataSetAndField(ComplexName: string; var DataSet: TDataSet;
  var Field: TField);

var
  DescriptionList: TStringList;

implementation

uses FR_Class, FR_Const, TypInfo;

procedure AddDescription(FunctionName, Category, Help: string; ParamNum: Integer);
const
  Parametri: array[1..3] of string = ('<X>', '<Y>', '<Z>');
var
  I: Integer;
  ParamStr: string;
begin
  ParamStr := '';
  for I := 1 to ParamNum do
    begin
      ParamStr := ParamStr + Parametri[I];
      if I <> ParamNum then
        ParamStr := ParamStr + ','
    end;
  DescriptionList.Add(FunctionName + '=' + Category + '|' + FunctionName + '(' + ParamStr + ') |' + Help);
end;

procedure RemoveQuotes(var s: string);
begin
  if (s[1] = '"') and (s[Length(s)] = '"') then
    s := Copy(s, 2, Length(s) - 2);
end;

procedure frReadMemo(Stream: TStream; l: TStrings);
var
  s: string[255];
  Y: string;
  I, J: Integer;
  B: Byte;
  W, X: Word;
  A: array[0..1023] of Char;
begin
  l.Clear;
  if frVersion < 26 then
    begin
      i := 1;
      repeat
        Stream.Read(b, 1);
        if (b = 13) or (b = 0) then
          begin
            s[0] := Chr(i - 1);
            if not ((b = 0) and (i = 1)) then l.Add(s);
            i := 1;
          end
        else
          if b <> 0 then
            begin
              s[i] := Chr(b);
              Inc(i);
            end;
      until b = 0;
    end
  else
    begin
      Stream.Read(X, 2);
      for I := 0 to X - 1 do
        begin
          Stream.Read(W, 2);
          Stream.Read(A, W);
          Y := '';
          for J := 0 to W - 1 do
            Y := Y + A[J];
          L.Add(Y);
        end;
    end;
end;

procedure frWriteMemo(Stream: TStream; L: TStrings);
var
  I: Integer;
  B: Word;
  A: array[0..1023] of Char;
begin
  B := L.Count;
  Stream.Write(B, 2);
  for I := 0 to L.Count - 1 do
    begin
      StrCopy(A, PChar(L[I]));
      B := Length(L[I]);
      if B > SizeOf(A) then B := SizeOf(A);
      Stream.Write(B, 2);
      Stream.Write(A, B);
    end;
end;

procedure frEnableControls(c: array of TControl; e: Boolean);
const
  Clr1: array[Boolean] of TColor = (clGrayText, clWindowText);
  Clr2: array[Boolean] of TColor = (clBtnFace, clWindow);
var
  i: Integer;
begin
  for i := Low(c) to High(c) do
    if c[i] is TLabel then
      with c[i] as TLabel do
        begin
          Font.Color := Clr1[e];
          Enabled := e;
        end
    else
      if c[i] is TEdit then
        with c[i] as TEdit do
          begin
            Color := Clr2[e];
            Enabled := e;
          end;
end;

function frGetDataSet(ComplexName: string): TDataSet;
var
  n: Integer;
  f: TComponent;
  s1, s2: string;
begin
  Result := nil;
  f := CurReport.Owner;
  n := Pos('.', ComplexName);
  if n = 0 then
    f := f.FindComponent(ComplexName)
  else
    begin
      s1 := Copy(ComplexName, 1, n - 1); // module name
      s2 := Copy(ComplexName, n + 1, 255); // database name
      f := FindGlobalComponent(s1);
      if f <> nil then f := f.FindComponent(s2);
    end;
  if f <> nil then
    Result := f as TDataSet
  else
    DoError(FRConst_ErrorLoadfrDataSet + ' ' + ComplexName);
end;

procedure frGetDataSetAndField(ComplexName: string; var DataSet: TDataSet;
  var Field: TField);
var
  n: Integer;
  f: TComponent;
  s1, s2, s3: string;
begin
  Field := nil;
  f := CurReport.Owner;
  n := Pos('.', ComplexName);
  if n <> 0 then
    begin
      s1 := Copy(ComplexName, 1, n - 1); // table name
      s2 := Copy(ComplexName, n + 1, 255); // field name
      if (s2[1] <> '"') and (Pos('.', s2) <> 0) then // module name present
        begin
          s3 := Copy(s2, Pos('.', s2) + 1, 255);
          s2 := Copy(s2, 1, Pos('.', s2) - 1);
          f := FindGlobalComponent(s1);
          if f <> nil then
            begin
              DataSet := f.FindComponent(s2) as TDataSet;
              RemoveQuotes(s3);
              if DataSet <> nil then Field := DataSet.FindField(s3);
            end;
        end
      else
        begin
          DataSet := f.FindComponent(s1) as TDataSet;
          RemoveQuotes(s2);
          if DataSet <> nil then
            Field := DataSet.FindField(s2)
          else
            begin
              f := FindGlobalComponent(s1);
              if f <> nil then
                DataSet := f.FindComponent(s2) as TDataSet
            end;
        end;
    end
  else
    if DataSet <> nil then
      begin
        RemoveQuotes(ComplexName);
        Field := DataSet.FindField(ComplexName);
      end;
end;

initialization
  DescriptionList := TStringList.Create;

finalization
  DescriptionList.Free;

end.

